#ifndef DETAIL_DOUBCONV_HH
#define DETAIL_DOUBCONV_HH

// ======================================================================
//
// $Id: DoubConv.hh,v 1.1 2006/02/08 15:25:00 mf Exp $
//
// ======================================================================


// ----------------------------------------------------------------------
// Prolog

// --- standard library facilities
#include <exception>
#include <string>
#include <vector>


namespace splines  {


// ======================================================================
//
// Contents
//
// ======================================================================


class DoubConvException;

class DoubConv;
class LongDoubConv;


// ======================================================================
//
// DoubConvException
//
// ======================================================================


class DoubConvException
  : public std::exception
{
public:
  DoubConvException( std::string const & s )
    : reason(s)
  { }

  ~DoubConvException() throw()
  { }

  virtual
    char const *
    what() const throw()
  { return reason.c_str(); }

private:
  std::string reason;

};  // DoubConvException


// ======================================================================
//
// DoubConv
//
// ======================================================================


class DoubConv
{
public:

  // dto2longs(d) returns (in a vector) two unsigned longs containing the
  // representation of its double input.  This is byte-ordering
  // independent, and depends for complete portability ONLY on adherance
  // to the IEEE 754 standard for 64-bit floating point representation.
  // The first unsigned long contains the high-order bits in IEEE; thus
  // 1.0 will always be 0x3FF00000, 00000000
  static std::vector<unsigned long> dto2longs(double d);

  // longs2double (v) returns a double containing the value represented by its
  // input, which must be a vector containing 2 unsigned longs.
  // The input is taken to be the representation according to
  // the IEEE 754 standard for a 64-bit floating point number, whose value
  // is returned as a double.  The byte-ordering of the double result is,
  // of course, tailored to the proper byte-ordering for the system.
  static double longs2double (std::vector<unsigned long> const & v);

  // dtox(d) returns a 16-character string containing the (zero-filled) hex
  // representation of its double input.  This is byte-ordering
  // independent, and depends for complete portability ONLY on adherance
  // to the IEEE 754 standard for 64-bit floating point representation.
  static std::string d2x(double d);

private:
  union DB8 {
    unsigned char b[8];
    double d;
  };
  static void fill_byte_order ();
  static bool byte_order_known;
  static int  byte_order[8];
    // Meaning of byte_order:  The first (high-order in IEEE 754) byte to
    // output (or the high-order byte of the first unsigned long)
    // is  of db.b[byte_order[0]].  Thus the index INTO byte_order
    // is a position in the IEEE representation of the double, and the value
    // of byte_order[k] is an offset in the memory representation of the
    // double.

};  // DoubConv


// ======================================================================
//
// LongDoubConv
//
// ======================================================================


class LongDoubConv
{
public:

  // longdto4longs(d) returns (in a vector) four unsigned longs containing the
  // representation of its long double input.  This is byte-ordering
  // independent, and depends for complete portability ONLY on the assumption
  // that the representation of long double involves a twos-complement
  // exponent and a mantissa (with or without an implied leading 1), and is
  // no longer than 16 bytes in length.   The portable representation will
  // always be as four longs (with the low-order 32-bits active in each), even
  // if the actual representation uses fewer bytes.
  //
  // The form of output will be a sign bit, followed by a **15-bit** exponent
  // (offset by 2^14-1) followed by 112 bits representing the bits to the right
  // of the binary point adjoined to an implied 1 to the left.  Thus
  // 1.0 will always be 0x3FFF0000, 00000000, 00000000, 00000000.  See
  // http://754r.ucbtest.org/ for a link to a pdf version of the draft
  // revised IEEE 754 standard.
  //
  // For platforms using non-IEEE formats, it is possible that various
  // denorms and/or NANs will not transfer portably, and certain that
  // numbers which are out of range on the target platform or contain
  // non-zero bits in low order words which exceed the byte length of long
  // double on the target platform will not be perfectly reproducible.

  static std::vector<unsigned long> longdto4longs(long double d);

  // longs2longdouble (v) returns a double containing the value represented
  // by its input, which must be a vector containing 4 unsigned longs.
  // The input is taken to be the representation according to
  // the IEEE 754 standard for a 128-bit floating point number, whose value
  // is returned as a long double.
  // The byte-ordering and content of the result is, of course,
  // tailored to the proper byte-ordering and content for the system.
  static long double longs2longdouble (std::vector<unsigned long> const & v);

  // longd2x(d) returns a 32-character string containing the (zero-filled) hex
  // representation of its double input.  This is system independent under
  // the same circumstances as for longdto4longs.
  static std::string longd2x(long double d);

private:
  union DB16 {
    unsigned char b[16];
    long double d;
  };
  static unsigned const bytes = sizeof(long double);
  static void fill_byte_order ();
  static bool byte_order_known;
  static int  mantissaN;    // Number of mantissa bits past implied leading 1
  static bool javaEncoding; // Expression as sum of two doubles
  static int  nSignBits;    // Number of sign bits (relevant for java encoding)
  static int  byte_order[16];
    // Meaning of byte_order:  The first (high-order in IEEE 754) byte to
    // output (or the high-order byte of the first unsigned long)
    // is  of db.b[byte_order[0]].  Thus the index INTO byte_order
    // is a position in the IEEE representation of the double, and the value
    // of byte_order[k] is an offset in the memory representation of the
    // double.

};  // LongDoubConv


// ----------------------------------------------------------------------
// Epilog

}  // namespace splines

#endif  // DETAIL_DOUBCONV_HH
