// Spline-1.cc
//
// Implementation for methods of Spline<1>
//

#include "splines/Spline.h"
#include "splines/Grid.h"
#include "splines/DeltaFunction.h"
#include "splines/GridSearchAnswer.h"
#include "splines/DoubConv.h"
#include <vector>
#include <algorithm>
#include <cassert>
#include <sstream>
#include <iomanip>
#include <cmath>
using std::setw;

//#define TRACE_S1
//#define TRACE_S2
//#define TRACE_S3
//#define TRACE_S4
//#define TRACE_S5

namespace splines {

struct WrapperFor_1_ArgFunction {
  WrapperFor_1_ArgFunction(double ff(std::vector<double> const & v)) 
  : f(ff) 
  , v(1) { }
  double operator() (double x1)  const { v[0] = x1; return f(v); }
  mutable std::vector<double> v;
  double (*f)(std::vector<double> const & v);
};

Spline<1>::Spline ( double f(double), const Grid<1> & g  ) 
  : grid(g)
  , specifiedByTable(false)
#ifdef EMBEDDING
  , embedding()
#endif
  , isDeltaFunction(false)
{
  captureValues ( f );
  computeDerivatives  ( );
}

Spline<1>::Spline ( double f(int x_index), const Grid<1> & g )   
  : grid(g)
  , specifiedByTable(true)
#ifdef EMBEDDING
  , embedding()
#endif
  , isDeltaFunction(false)
{
  captureValues ( f );
  computeDerivatives  ( );
}

Spline<1>::Spline ( std::vector<double> const & suppliedValues, 
                    const Grid<1> & g )
  : grid(g)
  , specifiedByTable(true)
#ifdef EMBEDDING
  , embedding()
#endif
  , isDeltaFunction(false)
{
  captureValues ( suppliedValues );
  computeDerivatives  ( );
}

Spline<1>::Spline ( const Grid<1> & g, std::vector<double> const & vals, 
                                       std::vector<double> const & derivs) 
  : grid(g)
  , values(vals)
  , derivatives(derivs)
#ifdef EMBEDDING
  , embedding()
#endif
  , isDeltaFunction(false)
{
}

Spline<1>::Spline ( double f(std::vector<double> const & v), const Grid<1> & g )
  : grid(g)
  , specifiedByTable(false)
#ifdef EMBEDDING
  , embedding()
#endif
  , isDeltaFunction(false)
{  
  captureValues ( f );
  computeDerivatives  ( ); 
}

Spline<1>::Spline ( double f(std::vector<int> const & indicies), 
		    const Grid<1> & g )
  : grid(g)
  , specifiedByTable(false)
#ifdef EMBEDDING
  , embedding()
#endif
  , isDeltaFunction(false)
{  
  captureValues ( f );
  computeDerivatives  ( ); 
}

Spline<1>::Spline ( DeltaFunction<1> const & d  )   
  : grid(d.x - 2.0*d.integral/d.h, d.x + (8.0/3.0)*d.integral/d.h , 7)
  , specifiedByTable(true)
  , embedding()
  , isDeltaFunction(true)
  , deltaFunctionLocation (d.x)
  , deltaFunctionStrength (d.integral)
{
  std::vector<double> vals(7);
  vals[0] = 0;
  vals[1] = 0;
  vals[2] = 0.25*d.h;
  vals[3] = d.h;
  vals[4] = 0.25*d.h;
  vals[5] = 0;
  vals[6] = 0;
  #ifdef TRACE_S3
  std::cout << "captureValues (vals) on grid\n";
  for (unsigned int i=0; i<7; ++i) {
    std::cout << grid[i] << "\n";
  }
  #endif
  captureValues ( vals );
  computeDerivatives  ( );
}
 
#ifdef EMBEDDING
Spline<1>::Spline ( double f(std::vector<double> const & v), 
		    const  Grid<1> & g,
  	   	    double outerF(std::vector<double> const & v ) )
  : grid(g)
  , specifiedByTable(true)
  , embedding( WrapperFor_1_ArgFunction(outerF) )  
  , isDeltaFunction(false)
{  
  captureValues ( f );
  computeDerivatives  ( ); 
}
				   
Spline<1>::Spline ( double f(std::vector<int> const & indicies), 
		    const  Grid<1> & g, 
  	   	    double outerF(std::vector<double> const & v ) )
  : grid(g)
  , specifiedByTable(true)
  , embedding( WrapperFor_1_ArgFunction(outerF) )  
  , isDeltaFunction(false)
{  
  captureValues ( f );
  computeDerivatives  ( ); 
}

Spline<1>::Spline ( const  Grid<1> & g, 
		    std::vector<double> const & vals, 
		    std::vector<double> const & derivs,
  	   	    double outerF(std::vector<double> const & v ) )
  : grid(g)
  , values(vals)
  , derivatives(derivs)
  , embedding( WrapperFor_1_ArgFunction(outerF) )  
  , isDeltaFunction(false)
{ 
}
#endif

void
Spline<1>::setTables (std::istream & is) {
  is >> grid;
  unsigned int N = grid.nPoints();
  std::vector<unsigned long> v(2);
  for (unsigned int i=0; i<N; ++i) {
    is >> v[0] >> v[1];
    values.push_back ( DoubConv::longs2double(v) );
  }
  for (unsigned int j=0; j<N; ++j) {
    is >> v[0] >> v[1];
    derivatives.push_back ( DoubConv::longs2double(v) );
  }
}

void 
Spline<1>::captureValues ( double f(double) ) {
  int n = grid.nPoints();
  values.clear(); values.reserve( n );
  double x;
  for (unsigned int index = 0; index < n; ++index) {
    x = grid[index];
    values.push_back ( f(x) );
  }
  specifiedByTable = false;
  const double smallFraction = .01;
  double xs = grid[0] * (1-smallFraction) + grid[1] * smallFraction;
  slope0 = (f(xs) - values[0])/(xs - grid[0]); 
  xs = grid[n-1] * (1-smallFraction) + grid[n-2] * smallFraction; 
  slopeN = (values[n-1] - f(xs))/(grid[n-1] - xs);  
}

void 
Spline<1>::captureValues(std::vector<double> const & suppliedValues) {
  int N = grid.nPoints();
  if (suppliedValues.size() != N) {
    throw Spline_table_exception(1,suppliedValues.size(),N);
  }
  values = suppliedValues;
  specifiedByTable = true;
  #ifdef TRACE_S3
  std::cout << "Captured values:\n";
  for (unsigned int i = 0; i<N; ++i) {
    std::cout << values[i] << "\n";
  }
  #endif
}

void 
Spline<1>::captureValues(double f(int x_index) ) {
  int N = grid.nPoints();
  values.clear(); values.reserve( N );
  for (int index = 0; index<N; ++index) {
    values.push_back ( f(index) );
  }
  specifiedByTable = true;
}

void 
Spline<1>::captureValues ( double f(std::vector<double> const &) ){
  int n = grid.nPoints();
  std::vector<double> v(1);
  values.clear(); values.reserve( n );
  for (unsigned int index = 0; index < n; ++index) {
    v[0] = grid[index];
    values.push_back ( f(v) );
  }
  specifiedByTable = false;
  const double smallFraction = .01;
  // These next steps compute derivative values to use at the endpoints.
  // Note, however, that since specifiedByTable = false, these captured
  // derivative values are going to be ignored in secondDerivatives, in
  // favor of using "natural" boundary conditions.
  // TODO - eliminate these four lines and re-test to see that it makes no 
  //        difference
  v[0] = grid[0] * (1-smallFraction) + grid[1] * smallFraction;
  slope0 = (f(v) - values[0])/(v[0] - grid[0]); 
  v[0] = grid[n-1] * (1-smallFraction) + grid[n-2] * smallFraction; 
  slopeN = (values[n-1] - f(v))/(grid[n-1] - v[0]);  
}

void 
Spline<1>::captureValues( double f(std::vector<int> const &) ){
  int N = grid.nPoints();
  std::vector<int> indicies(1);
  values.clear(); values.reserve( N );
  for (int index = 0; index<N; ++index) {
    indicies[0] = index;
    values.push_back ( f(indicies) );
  }
  specifiedByTable = true;
}

void 
Spline<1>::computeDerivatives ( ) {

  std::vector<double> x  (grid.points());
  int N = x.size();
  std::vector<double> y2 (N);
  if (specifiedByTable) {
    y2 = secondDerivatives (x, values, 0, 0, 0, 0 );
  } else {
    double s0 = 3 * ( (values[1]-values[0])/(x[1]-x[0]) - slope0 ) 
							/ (x[1]-x[0]);
    double sN = 3 * ( slopeN - (values[N-1]-values[N-2])/(x[N-1]-x[N-2]) ) 
							/ (x[1]-x[0]);
    y2 = secondDerivatives (x, values, .5, s0, .5, sN ); 
  }
  
  // apply 3.3.5, with A =1 and B = 0 for all points except the last:

  #ifdef TRACE_S3
  std::cout << "derivatives: \n";     
  #endif

  for (int i=0; i<N-1; ++i) {
    #ifdef TRACE_S3
    std::cout << i << ":   " <<  (values[i+1]-values[i])/(x[i+1]-x[i])
			           - (x[i+1]-x[i]) * (2*y2[i] + y2[i+1])/6.0 
              << "\n";  
    #endif
    derivatives.push_back (   (values[i+1]-values[i])/(x[i+1]-x[i])
			           - (x[i+1]-x[i]) * (2*y2[i] + y2[i+1])/6.0  );
  }

  // For the last point, use A = 0 and B = 1 in the last interval:
  derivatives.push_back (   (values[N-1]-values[N-2])/(x[N-1]-x[N-2])
			         + (x[N-1]-x[N-2]) * (y2[N-2] + 2*y2[N-1])/6.0);
  
  #ifdef TRACE_S3
  std::cout << N-1 << ":   " <<  (values[N-1]-values[N-2])/(x[N-1]-x[N-2])
			         + (x[N-1]-x[N-2]) * (y2[N-2] + 2*y2[N-1])/6.0 
              << "\n";  
  #endif
  
  #ifdef TRACE_S2
  std::cout << 
  "  i         x         values   derivative    second derivative\n";
  for (int i=0; i<N; ++i) {
    std::cout << setw(6) << i 
	  << "  " << setw(10) << x[i] << "  " 
	  << setw(10) << values[i] << "  "
	  << setw(10) << derivatives[i] << "  "
	  << setw(10) << y2[i] << "\n";
  }
  #endif
}


void
Spline<1>::swap (Spline<1> & s) {
  grid.swap(s.grid);
  values.swap(s.values);
  derivatives.swap(s.derivatives);
  std::swap (specifiedByTable, s.specifiedByTable);
  std::swap (embedding, s.embedding);
  std::swap (slope0, s.slope0);
  std::swap (slopeN, s.slopeN);
  std::swap (isDeltaFunction, s.isDeltaFunction);
  std::swap (deltaFunctionLocation, s.deltaFunctionLocation);
  std::swap (deltaFunctionStrength, s.deltaFunctionStrength);
}


std::vector<double> 
Spline<1>::
secondDerivatives (const std::vector<double> & x, 
		   const std::vector<double> & y,
		   double a0, double s0,
		   double bN, double sN) {

  int N = x.size();
  std::vector<double> y2 (N);

  // We will be solving M y2 = s, where M is a tri-diagonal matrix.
  // $$$ See Mathematical Details 1.2 $$$ 

  std::vector<double> m (N);  // Diagonal elements of M 
  std::vector<double> a (N);  // Elements of M above the diagonal (0 - N-2)
  std::vector<double> b (N);  // Elements of M below the diagonal (1 - N-1)
  std::vector<double> s (N);  
  // Optimization opportunity:
  //
  // This can be done with only two vector of temp storage space, s and a.
  // For code clarity, we use these four instead.

  // Step 1: Form M and s 

  if ( a0 != 0 ) {	// derivative Boundary condition at 0
    assert (a0 == .5);
  } else {
    s0 = 0;
  }

  if ( bN != 0 ) {	// derivative Boundary condition at N-1
    assert (bN == .5);
  } else {
    sN = 0;
  }

  m[0] = 1;
  a[0] = a0;
  s[0] = s0;
  #ifdef TRACE_S3
  std::cout << "            m[0] = " << m[0] 
            << "  a[0] = " << a[0]
            << "  s[0] = " << s[0] << "\n";
  #endif

  int i;
  for ( i = 1; i < N-1; ++i ) {
    b[i] = ( x[i]   - x[i-1] ) /6.0;
    m[i] = ( x[i+1] - x[i-1] ) /3.0;
    a[i] = ( x[i+1] - x[i]   ) /6.0;
    s[i] = (y[i+1] - y[i])/(x[i+1] - x[i]) - (y[i] - y[i-1])/(x[i] - x[i-1]); 
    #ifdef TRACE_S3
    std::cout << "  b[" <<i<<"] = " << b[i] 
              << "  m[" <<i<<"] = " << m[i] 
              << "  a[" <<i<<"] = " << a[i]
              << "  s[" <<i<<"] = " << s[i] << "\n";
    #endif
  }

  b[N-1] = bN;
  m[N-1] = 1;
  s[N-1] = sN;

  #ifdef TRACE_S3
  std::cout << "  b[" <<N-1<<"] = " << b[N-1]  
            << "  m[" <<N-1<<"] = " << m[N-1]
            << "  s[" <<N-1<<"] = " << s[N-1] << "\n";
  #endif

  // Step 2: Eliminate the lower diagonal elements; m[i] becomes implied 1's

  double multiplier;
  for ( i = 0; i < N-1; ++i ) {
    multiplier = 1.0/m[i];
    a[i] *= multiplier;
    s[i] *= multiplier;
    m[i+1] -= b[i+1] * a[i];
    s[i+1] -= b[i+1] * s[i];
  }
  multiplier = 1.0/m[i];
  s[N-1] /= m[N-1];

  // Step 3: Backsolve (eliminate the upper diagonal a[i])

  for ( i = N-1; i > 0; --i ) {
    s[i-1] -= a[i-1] * s[i];
  }
  #ifdef TRACE_S3
  std::cout << "Second derivatives vector: \n";
  for (i=0; i<N; ++i) {
    std::cout << i << ":  " << s[i] << "\n";
  }
  #endif
  return s;
}


double
Spline<1>::integrate(double a, double b) const
{
  if (isDeltaFunction) {
    if ( (a <= deltaFunctionLocation) && (b >= deltaFunctionLocation) ) {
      return deltaFunctionStrength;
    } else {
      return 0;
    }
  }
  GridSearchAnswer ga;
  grid.search (a, ga);
  GridSearchAnswer gb;
  grid.search (b, gb);
  double x = ga.x;
  double x2 = x*x;
  double x3 = x*x2;
  double x4 = x2*x2;
  int index = ga.index;
  #ifdef TRACE_S4
    std::cout << "a = " << a << " index = " << index << " x = " << x << "\n";
  #endif
  double integral = ga.h * (
           values[index] * (.5 - .5*x4 + x3 - x)
         + values[index+1]*(.5 + .5*x4 - x3)
         + ga.h*derivatives[index]*(1 - 3*x4 + 8*x3 - 6*x2)/12.0
         + ga.h*derivatives[index+1]*(-1 - 3*x4 + 4*x3)/12.0 );
  #ifdef TRACE_S4
    std::cout << "at " << a
              << " values are " << values[index] << ", " << values[index+1]  
              << "\nderivs are " << derivatives[index] << ", " 
              << derivatives[index+1]  
              << " h = " << ga.h << "\n"
              << "integral thus far is " << integral << "\n";
  #endif
  ++index;
  while (index < gb.index) {  
     double h = grid[index+1] - grid[index];
     integral += h *  ( 
          (values[index] + values[index+1]) * .5 
        + (derivatives[index] - derivatives[index+1]) * (h/12.0) );
    #ifdef TRACE_S4
    std::cout << " at " << grid[index] 
              << " values are " << values[index] << ", " << values[index+1]  
              << "\nderivs are " << derivatives[index] << ", " 
              << derivatives[index+1]  
              << " h = " << h << "\n"
              << "integral thus far is " << integral << "\n";
    #endif
    #ifdef TRACE_S5
    std::cout << "    Integral contributions:  \n" 
              << "        " << h *  values[index] *.5 << "\n"
              << "        " << h *  values[index+1] *.5<< "\n"
              << "        " << h*h * derivatives[index] / 12 << "\n"
              << "        " << -h*h * derivatives[index+1] / 12 << "\n"
              << "======= " << h *  ( 
                       (values[index] + values[index+1]) * .5 
                       + (derivatives[index]- derivatives[index+1]) * (h/12.0) )
              << "\n";
    #endif 
    ++index;
  }
  x = gb.x;
  x2 = x*x;
  x3 = x*x2;
  x4 = x2*x2;
  integral += gb.h * (
          values[index] * (.5*x4 - x3 + x)
        + values[index+1]*(-.5*x4 + x3)
        + gb.h*derivatives[index]*(3*x4 - 8*x3 + 6*x2)/12.0
        + gb.h*derivatives[index+1]*(3*x4-4*x3)/12.0 );
  #ifdef TRACE_S4
    std::cout << " final x is " << x 
              << " values are " << values[index] << ", " << values[index+1]  
              << "\nderivs are " << derivatives[index] << ", " 
              << derivatives[index+1]  
              << " h = " << gb.h << "\n"
              << "integral is " << integral << "\n";
  #endif
   return integral;
} // integrate

Spline<1>
Spline<1>::indefiniteIntegral(double lowerLimit) const
{
  if (isDeltaFunction) {
    // The indefinite integral of a delta function is a theta fucntion:
    // We will construct one explicitly.
    std::vector<double> vals(7);
    vals[0] = 0;
    vals[1] = 0;
    vals[2] = deltaFunctionStrength/12;
    vals[3] = 0.5 * deltaFunctionStrength;
    vals[4] = (11.0/12.0)*deltaFunctionStrength;
    vals[5] = deltaFunctionStrength;
    vals[6] = deltaFunctionStrength;
    #ifdef TRACE_S3   
    std::cout << "Delta function spline indefiniteIntegral \n";
    for (int i=0; i<grid.nPoints(); ++i) {
      std::cout << grid[i] << " ---- " << vals[i] << "\n";
    } 
    #endif
    return Spline<1>(vals, grid); 
  } 
  #ifdef TRACE_S5
  int oldPrecision = std::cout.precision(12);
  std::cout << "setting precision to 12: " << 1.23456789123456 << "\n";
  #endif 
  std::vector<double> integralValues(grid.nPoints());
  // First, integrate from the start of the grid to the lower limit,
  // placing tentative answers into the answer vector.
  int index = 0;
  GridSearchAnswer gb;
  grid.search (lowerLimit, gb);
  double integral = 0;
  integralValues[index] = integral; 
  while (index < gb.index) {  
     double h = grid[index+1] - grid[index];
     integral += h *  ( 
          (values[index] + values[index+1]) * .5 
        + (derivatives[index] - derivatives[index+1]) * (h/12.0) );
    #ifdef TRACE_S4
    std::cout << " at " << grid[index] 
              << " values are " << values[index] << ", " << values[index+1]  
              << "\nderivs are " << derivatives[index] << ", " 
              << derivatives[index+1]  
              << " h = " << h << "\n"
              << "integral thus far is " << integral << "\n";
    #endif
    #ifdef TRACE_S5
    std::cout << "    Integral contributions:  \n" 
              << "        " << h *  values[index] *.5 << "\n"
              << "        " << h *  values[index+1]*.5 << "\n"
              << "        " << h*h * derivatives[index] / 12 << "\n"
              << "        " << -h*h * derivatives[index+1] / 12 << "\n"
              << "======= " << h *  ( 
                       (values[index] + values[index+1]) * .5 
                       + (derivatives[index] - derivatives[index+1]) * (h/12.0) )
              << "\n";
    #endif 
    ++index;
    integralValues[index] = integral;
  }
  // Now do the last little interval up to the lower limit
  double x = gb.x;
  double x2 = x*x;
  double x3 = x*x2;
  double x4 = x2*x2;
  integral += gb.h * (
          values[index] * (.5*x4 - x3 + x)
        + values[index+1]*(-.5*x4 + x3)
        + gb.h*derivatives[index]*(3*x4 - 8*x3 + 6*x2)/12.0
        + gb.h*derivatives[index+1]*(3*x4-4*x3)/12.0 );
    #ifdef TRACE_S5
    std::cout << "    Integral contributions:  \n" 
              << "    " << gb.h *  values[index] *(.5*x4 - x3 + x) << "\n"
              << "    " << gb.h *  values[index+1]*(-.5*x4 + x3) << "\n"
              << "    " << gb.h*gb.h  * 
                    derivatives[index]*(3*x4 - 8*x3 + 6*x2)/12.0 << "\n"
              << "    " << -gb.h*gb.h * 
                    derivatives[index+1]*(3*x4-4*x3)/12.0 << "\n"
              << "=== " << gb.h * (
          values[index] * (.5*x4 - x3 + x)
        + values[index+1]*(-.5*x4 + x3)
        + gb.h*derivatives[index]*(3*x4 - 8*x3 + 6*x2)/12.0
        + gb.h*derivatives[index+1]*(3*x4-4*x3)/12.0 )           << "\n";
    #endif 
  #ifdef TRACE_S4
    std::cout << "final x before lower limit is " << x 
              << " values are " << values[index] << ", " << values[index+1]  
              << "\nderivs are " << derivatives[index] << ", " 
              << derivatives[index+1]  
              << " h = " << gb.h << "\n"
              << "integral is " << integral << "\n";
  #endif
  double integralToLowerLimit = integral;
  // Now correct the answer vector by subtracting off the integral to LL
  for (unsigned int i=0; i < index+1; ++i) {
    integralValues[i] -= integralToLowerLimit;
  }
  // Now complete the integral, till the end of the grid.
  integral = 0;  // because at x=LL, integral is indeed supposed to be zero
  // Start in the slot that LowerLimit is in, using int (x,1) formula
  #ifdef TRACE_S4
    std::cout << "a = " << lowerLimit << " index = " << index 
              << " x = " << x << "\n";
  #endif
  integral = gb.h * (
           values[index] * (.5 - .5*x4 + x3 - x)
         + values[index+1]*(.5 + .5*x4 - x3)
         + gb.h*derivatives[index]*(1 - 3*x4 + 8*x3 - 6*x2)/12.0
         + gb.h*derivatives[index+1]*(-1 - 3*x4 + 4*x3)/12.0 );
  #ifdef TRACE_S5
  std::cout << "    Integral contributions:  \n" 
            << "    " << gb.h *  values[index] *(.5 - .5*x4 + x3 - x) << "\n"
            << "    " << gb.h *  values[index+1]*(.5 + .5*x4 - x3) << "\n"
            << "    " << gb.h*gb.h  * 
                  derivatives[index]*(1 - 3*x4 + 8*x3 - 6*x2)/12.0 << "\n"
            << "    " << -gb.h*gb.h * 
                  derivatives[index+1]*(-1 - 3*x4 + 4*x3)/12.0/12.0 << "\n"
            << "=== " << gb.h * (
           values[index] * (.5 - .5*x4 + x3 - x)
         + values[index+1]*(.5 + .5*x4 - x3)
         + gb.h*derivatives[index]*(1 - 3*x4 + 8*x3 - 6*x2)/12.0
         + gb.h*derivatives[index+1]*(-1 - 3*x4 + 4*x3)/12.0 )       << "\n";
  #endif 
  #ifdef TRACE_S4
    std::cout << "at " << grid[index]
              << " values are " << values[index] << ", " << values[index+1]  
              << "\nderivs are " << derivatives[index] << ", " 
              << derivatives[index+1]  
              << " h = " << gb.h << "\n"
              << "integral thus far is " << integral << "\n";
  #endif
  ++index;
  integralValues[index] = integral;
  while (index < grid.nPoints()-1) {  
     double h = grid[index+1] - grid[index];
     integral += h *  ( 
          (values[index] + values[index+1]) * .5 
        + (derivatives[index] - derivatives[index+1]) * (h/12.0) );
    #ifdef TRACE_S4
    std::cout << " at " << grid[index] 
              << " values are " << values[index] << ", " << values[index+1]  
              << "\nderivs are " << derivatives[index] << ", " 
              << derivatives[index+1]  
              << " h = " << h << "\n"
              << "integral thus far is " << integral << "\n";
    #endif
    #ifdef TRACE_S5
    std::cout << "    Integral contributions:  \n" 
              << "        " << h *  values[index] *.5 << "\n"
              << "        " << h *  values[index+1] << "\n"
              << "        " << h*h * derivatives[index] / 12 << "\n"
              << "        " << h*h * derivatives[index+1] / 12 << "\n"
              << "======= " << h *  ( 
                       (values[index] + values[index+1]) * .5 
                       + (derivatives[index]+ derivatives[index+1]) * (h/12.0) )
              << "\n";
    #endif     
    ++index;
    integralValues[index] = integral;
  }
  // (Unlike the definite integral case, there is no follow up for the last
  //  grid space that contains the endpoint b. )

  // Finally, form the spline from the computed integral values/
  // TODO -- we can slightly improve accuracy by using a spline ctor
  //         that takes the explicit slopes at the endipoints, because
  //         we know the values of the original spline, are the slopes of the
  //         integral.  This should not make a big difference, and there is
  //         some merit to the natural boundary conditions (2nd deriv = 0).
#ifdef TRACE_S5
  std::cout << "integralValues \n";
  for (unsigned int i = 0; i < grid.nPoints(); ++i) {
    std::cout << i << ":  "  << grid[i]  << "  " << integralValues[i] << "\n";
  }
#endif
#ifdef TRACE_S5
  std::cout.precision(oldPrecision);
#endif
  
  return Spline<1> (integralValues, grid);

} // indefiniteIntegral

double
Spline<1>::maximum() const
{
  // This will return the maximimum **within the grid**
  
  // TODO - This code is the safest and cleanest method - there are
  //        probably mathematical tricks to speed up the algorithm
  
  double xm = grid[0];
  double maxThusFar = f(xm);
  for (int i = 0; i < grid.nPoints()-1; ++i) {
    double endpoint  = grid[i+1];
    double endpointF = f(endpoint);
    #ifdef TRACE_S5
      std::cout << "max: xm = " << xm << " mtf = " << maxThusFar
                << " endpt = " << endpoint << " F = " << endpointF << "\n";
    #endif 
    if (endpointF > maxThusFar) {
      xm = endpoint;
      maxThusFar = endpointF;
      #ifdef TRACE_S5
        std::cout << "     xm = " << xm << " mtf = " << maxThusFar << "\n";
      #endif 
    }
    // Now check if there is a better extremum within the interval:
    const double h = grid[i+1] - grid[i];
    // Construct the cubic in this range: p*x^3 + q*x^2 + r*x + s
    const double f0 = values[i];
    const double f1 = values[i+1];
    const double g0 = h*derivatives[i];
    const double g1 = h*derivatives[i+1];
    double p = 2*f0 - 2*f1 + g0 + g1;
    double q = 3*f1 - 3*f0 -2*g0 -g1;
    double r = g0;
    double s = f0;
    // Does the derivative have any zeros? The derivative is a*x^2 + b*x + c
    double a = 3*p;
    double b = 2*q;
    double c = r;
    double D = b*b - 4*a*c;
    if (D < 0) continue;  // no zero's of derivative --> monotonic in interval
    // Are any solutions (extrema) in the interval?
    double sqrtD = std::sqrt(D);
    double denomRecip = 1.0/(2.0*a);
    double x1 = (-b - sqrtD)*denomRecip; 
    double x2 = (-b + sqrtD)*denomRecip; 
    if ( (0 < x1) && (x1 < 1) ) {
      #ifdef TRACE_S5
        std::cout << "     local extremum at x1 = " << x1
                  << " which is x = " << grid[i] + h*x1 
                  << " f(x) = " << f(grid[i] + h*x1) <<  "\n";
      #endif 
      // test if this local extremum improves matters
      double v =  p*x1*x1*x1 + q*x1*x1 + r*x1 + s;
      if ( v > maxThusFar ) {
        xm = grid[i] + h*x1;
        maxThusFar = v;
        #ifdef TRACE_S5
          std::cout << "***  xm = " << xm << " mtf = " << maxThusFar << "\n";
        #endif 
      }
    }
    if ( (0 < x2) && (x2 < 1) ) {
      #ifdef TRACE_S5
        std::cout << "     local extremum at x2 = " << x2
                  << " which is x = " << grid[i] + h*x2 
                  << " f(x) = " << f(grid[i] + h*x2) <<  "\n";
      #endif 
      double v =  p*x2*x2*x2 + q*x2*x2 + r*x2 + s;
      if ( v > maxThusFar ) {
        xm = grid[i] + h*x2;
        maxThusFar = v;
        #ifdef TRACE_S5
          std::cout << "***  xm = " << xm << " mtf = " << maxThusFar << "\n";
        #endif 
      }
    }
  } // end of the for loop over grid intervals
  return xm;
} // maximum

double
Spline<1>::minimum() const
{
  // This will return the minimimum **within the grid**
  double xm = grid[0];
  double minThusFar = f(xm);
  for (int i = 0; i < grid.nPoints()-1; ++i) {
    double endpoint  = grid[i+1];
    double endpointF = f(endpoint);
    if (endpointF < minThusFar) {
      xm = endpoint;
      minThusFar = endpointF;
    }
    // Now check if there is a better extremum within the interval:
    const double h = grid[i+1] - grid[i];
    // Construct the cubic in this range: p*x^3 + q*x^2 + r*x + s
    const double f0 = values[i];
    const double f1 = values[i+1];
    const double g0 = h*derivatives[i];
    const double g1 = h*derivatives[i+1];
    double p = 2*f0 - 2*f1 + g0 + g1;
    double q = 3*f1 - 3*f0 -2*g0 -g1;
    double r = g0;
    double s = f0;
    // Does the derivative have any zeros? The derivative is a*x^2 + b*x + c
    double a = 3*p;
    double b = 2*q;
    double c = r;
    double D = b*b - 4*a*c;
    if (D < 0) break;  // no zero's of derivative --> monotonic in interval
    // Are any solutions (extrema) in the interval?
    double sqrtD = std::sqrt(D);
    double denomRecip = 1.0/(2.0*a);
    double x1 = (-b - sqrtD)*denomRecip; 
    double x2 = (-b + sqrtD)*denomRecip; 
    if ( (0 < x1) && (x1 < 1) ) {
      // test if this local extremum improves matters
      double v =  p*x1*x1*x1 + q*x1*x1 + r*x1 + s;
      if ( v < minThusFar ) {
        xm = grid[i] + h*x1;
        minThusFar = v;
      }
    }
    if ( (0 < x2) && (x2 < 1) ) {
      double v =  p*x2*x2*x2 + q*x2*x2 + r*x2 + s;
      if ( v < minThusFar ) {
        xm = grid[i] + h*x2;
        minThusFar = v;
      }
    }
  } // end of the for loop over grid intervals
  return xm;
} // minimum


double 
Spline<1>::operator() ( double xval ) const {

  // $$$ See Mathematical Details 1.1 $$$ 

  // Find the interval that xval lies within.  
  // $$$ For out-of-range cases, see Mathematical Details 1.4 $$$

  GridSearchAnswer a;
  bool interior = grid.search(xval, a);
				#ifdef TRACE_S1
				static int ntraces=0;
				if (ntraces < 10) { 
				std::cout << "Spline (" << xval << "): \n" 
				<< "x = " << a.x << " h = " << a.h 
				<< " a.index = " << a.index 
				<< "\nvalues " 	<<  values[a.index] 
				<<  ", " << values[a.index+1] 
				<< " grads  " <<  derivatives[a.index] << ", " 
				<< derivatives[a.index+1] << "\n";
				}
				#endif

  // Compute the f0, f1, g0 and g1 functions. 
  // $$$ See Mathematical Details 1.5 $$$

//  if ( (!interior) && embedding ) { 
//    return embedding(xval); 
//  }
  
#ifdef EMBEDDING
  if (!interior) {
    if (embedding) { 
      return embedding(xval);
    } 
  }
#endif
  
  double x = a.x;
  double oneMinusX = 1.0 - x;

  double spline;

  double x2 = x*x;
  double oneMinusX2 = oneMinusX * oneMinusX;
  double h = a.h;

  double f0 = (2. * x + 1.) * oneMinusX2;
  double f1 = 1.0 - f0;				// same as (3. - 2. * x) * x2;
  double g0 =   h * x  * oneMinusX2;
  double g1 =  -h * x2 * oneMinusX; 
				#ifdef TRACE_S1
				if (ntraces < 10) { ntraces++;
				std::cout << " f0 = " << f0 
		  			  << " f1 = " << f1
		  			  << " g0 = " << g0
		  			  << " g1 = " << g1 << "\n";
				}
				#endif 

  spline =   f0 * values[a.index]       + f1 * values[a.index+1]
	   + g0 * derivatives[a.index]  + g1 * derivatives[a.index+1] ;

  return spline;
 
} /* operator() */

double 
Spline<1>::operator() ( const std::vector<double> & v ) const {
  if ( v.size() != 1 ) throw  Spline_argument_exception(1,v.size());
  return operator() ( v[0] );
}

void 
Spline<1>::save (std::vector<double> & values_output, 
		 std::vector<double> & derivatives_output) const {
  values_output      = values;
  derivatives_output = derivatives;
}		 

bool
Spline<1>::operator==( Spline<1> const & rhs ) const {
  if (grid != rhs.grid) return false;
  unsigned int N = grid.nPoints();
  for (unsigned int i = 0;  i < N; ++i) {
    if (values[i]      != rhs.values[i]     ) return false;
    if (derivatives[i] != rhs.derivatives[i]) return false;
  }
  return true; 
}

std::ostream &
Spline<1>::put (std::ostream & os) const {
  os << grid;
  os << "\n";
  unsigned int N = grid.nPoints();
  std::vector<unsigned long> v(2);
  for (unsigned int i=0; i<N; ++i) {
    v = DoubConv::dto2longs(values[i]);
    os << v[0] << " " << v[1] << "\n";
  }
  os << "\n";
  for (unsigned int j=0; j<N; ++j) {
    v = DoubConv::dto2longs(derivatives[j]);
    os << v[0] << " " << v[1] << "\n";
  }
  return os;  
}

// =================
// Exception methods
// =================

const char* 
Spline_argument_exception::what() const throw() {
  std::ostringstream os;
  os << "Spline<" << d 
     << ">::operator()(vector):\n           requires v.size() = " << d
     << " v supplied is of size " << n;
  return os.str().c_str();
}

const char* 
Spline_table_exception::what() const throw() {
  std::ostringstream os;
  os << "Spline<" << d 
     << ">::ctor from vector of supplied values was supplied " << v
     << " values, but the grid has " << n << " points"; 
  return os.str().c_str();
}


} // end of namespace splines
